/*
linphone
Copyright (C) 2000  Simon MORLAT (simon.morlat@free.fr)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef LINPHONEMAIN_H
#define LINPHONEMAIN_H

#include <stdio.h>
#include <glib.h>

#include "../osipua/src/osipua.h"
#include "../osipua/src/sdphandler.h"
#include "../mediastreamer/mediastream.h"

/* version number in hex*/
#define INT_VERSION 0x000400  /*0.4.0*/

/* max number of network interfaces*/
#define MAX_IF 5

/* relative path where is stored local ring*/
#define LOCAL_RING "ring.raw"
/* same for remote ring (ringback)*/
#define REMOTE_RING "ringback.raw"

/* the default expire time sent in register requests*/
#define REG_EXPIRES 15*60


typedef struct registrar_config
{
	char *registrar;
	char *passwd;
	char *addr_of_rec;
	guint expires;
	guint32 use_registrar:1;
	guint32 as_proxy:1;
	guint32 as_redirect:1;
	guint32 as_outbound:1;
	guint32 padding:28;
} registrar_config_t;

typedef struct sip_config
{
	char *contact;
	char *username;
	char *hostname;
	registrar_config_t reg_conf;
	FILE *sip_dbg;
	guint16 sip_port;
} sip_config_t;

typedef struct rtp_config
{
	FILE *rtp_dbg;
	guint16 audio_rtp_port;
	guint16 video_rtp_port;
	gint jitt_comp;  /*jitter compensation*/
}rtp_config_t;

typedef struct net_config
{
	char if_names[MAX_IF][20];
	char if_addr[MAX_IF][20];
	char *local_addr;  /* the used address that is one somewhere above, but in dot notation*/
	char *if_name;   		/* the used interface name */
	double bandwidth; 		/* the bandwidth in bit/s of the selected interface*/ /*unused*/
	char con_type;			/* type of connection 28k 56k 128k adsl ...*/
	char if_sel;				/* index of the interface used*/
	char if_count;   /* number of interfaces found*/
}net_config_t;


typedef struct sound_config
{
	char *snd_name;   /* name of sound card*/ /*unused*/
	char *snd_dev;    /* /dev/dsp */			/*unused*/
	gint dev_id;		/* sound device id */
	gint latency;	/* latency in samples of the current used sound device */
	char rec_lev;
	char play_lev;
	char source;
	char pad;
	char *local_ring;
	char *remote_ring;
	int autokill:1;  /* 1= automatically free access to sound device by killing artsd and esd*/
	int driver_mode:2; /* 0=oss 1=alsa-0.5.x 2=alsa-0.9.x (for later use) */
} sound_config_t;

typedef struct codecs_config
{
	GList *audio_codecs;  /* list of audio codecs in order of preference*/
	GList *video_codecs;	/* for later use*/
	GList *codecs;		/*list of codecs, in any order*/
}codecs_config_t;

typedef struct gui_config
{
	GList *address_list;
	int is_daemon;
	int is_applet;
	guint timer_id;  /* the timer id for registration */
}gui_config_t;

/* this is the main class for linphone, can be used statically*/

typedef struct _LinphoneMain
{
	net_config_t net_conf;
	sip_config_t sip_conf;
	rtp_config_t rtp_conf;
	sound_config_t sound_conf;
	codecs_config_t codecs_conf;
	gui_config_t gui_conf;
	OsipUA *ua;			/* the osip user agent*/
	OsipDialog *call;   /* the current call, NULL if no call is running*/
	BodyHandler *sdph;	/* the sdp handler */
	AudioStream *audiostream;
	RingStream *ringstream;
}LinphoneMain;

/* table of different bandwidth for different connection type*/
extern const double bandwidths[5];


/* init object */
void linphone_main_init(LinphoneMain *lpm);

/* if there is no config file, use this functions to set default parameters*/
int linphone_main_first_startup(LinphoneMain *lpm);

/* this functions loads linphone configuration from a config file. If the file does not exists, returns -1.*/
int linphone_main_read_config(LinphoneMain *lpm);

/* save current linphone configuration*/
int linphone_main_save_config(LinphoneMain *lpm);

/* call this when the order of codec has changed*/
void linphone_main_apply_codec_change(LinphoneMain *lpm);

/* test if codecs are usable compared to bandwidth.*/
void linphone_main_codecs_check(LinphoneMain *lpm);

/*set a timer for registration if needed*/
void linphone_main_set_registration_timer(LinphoneMain *lpm);

/*remove the registration timer if present*/
void linphone_main_remove_registration_timer(LinphoneMain *lpm);

/* create the sip user agent and run it*/
int linphone_main_start(LinphoneMain *lpm);

/* stop all processes properly*/
void linphone_main_stop(LinphoneMain *lpm);

/* perform lasts things before leaving */
void linphone_main_shutdown(LinphoneMain *lpm,int (*cb)(OsipDialog* a,transaction_t*,sip_t *sip,void * b) );

/* final exit: stops all threads for gracefull shutdown*/
void linphone_main_exit(LinphoneMain *lpm);

/* set ua identity using linphone config */
int linphone_main_set_contact(LinphoneMain *lpm);

/* get the codec_info_t for codec implementing profile */
MSCodecInfo * linphone_main_get_codec(LinphoneMain *lpm, gint profile);

/***************************************************************************************************************/
/* private functions																																													*/
/**************************************************************************************************************/

/* retrieves interface names and address excluding local loopback. Returns the number of availlable interfaces found.*/
int retrieve_if(LinphoneMain *lpm);

/* kill esd and artsd if they are running, in order to free the sound device access*/
int kill_sound_daemons();
void restore_sound_daemons();



/*       LOCK FILE UTILITIES          */

/* put a lock file in /tmp. this is called when linphone runs as a daemon*/
int set_lock_file();

/* looks if there is a lock file. If presents return its content (the pid of the already running linphone), if not found, returns -1*/
int get_lock_file();

/* remove the lock file if it was set*/
int remove_lock_file();



#endif

