/*
  The osipua library is a library based on oSIP that implements CallLeg and User Agent
  level.
  Copyright (C) 2001  Simon MORLAT simon.morlat@free.fr
  											Aymeric MOIZARD jack@atosc.org
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "osipmanager.h"
#include "osipua.h"
#include "utils.h"
#include "uatransaction.h"
#include "resolver.h"
#include "digcalc.h"

list_t callleg_list = { 0, NULL };

void
osip_dialog_init (OsipDialog * new_call, OsipUA * ua)
{
	contact_t *ctt;

	memset (new_call, 0, sizeof (OsipDialog));
	/*set status for this new call */
	new_call->status = DIALOG_NEW;
	/* init transaction list */
	list_init (&new_call->incoming_transactions);
	list_init (&new_call->outgoing_transactions);
	/* set the ua reference on the newly created call-leg */
	new_call->ua = ua;
	/* add the call-leg to the list of call-legs handled by the stack */
	list_add (&callleg_list, (void *) new_call, -1);
	/* add the call-leg to the list of call-legs managed by the ua */
	list_add (&ua->call_list, (void *) new_call, -1);
	ctt = (contact_t *) list_get (ua->alias, 0);

	if (ctt != NULL)
	{
		from_clone ((from_t *) ctt, &new_call->from);
	}
	else
	{
		from_clone ((from_t *) ua->contact, &new_call->from);
	}


	/* BKO: initialize empty route list */
	list_init (&new_call->route_list);
	/* set from tag */
	from_tag_add (new_call->from, sgetcopy (ua->fromtag));
	/* set default registrar if given */
	if (ua->registrar != NULL)
	{
		url_clone (ua->registrar, &new_call->registrar);
	}
	/* init the list of body handlers */
	list_init (&new_call->body_contexts);
	ua->dialog_count++;
}

/* this is the init function that initialize the call-leg if it is created
locally: callid, ceq... are choosed randomly */
void
osip_dialog_init2 (OsipDialog * new_call, OsipUA * ua)
{
	char *number;

	/* set up callid */
	call_id_init (&new_call->callid);
	call_id_sethost (new_call->callid, sgetcopy (ua->ua_ip4addr));
	number = smalloc (17);
	snprintf (number, 15, "%i", rand ());
	call_id_setnumber (new_call->callid, number);
	/* set to ?? REVISIT */
	to_init (&new_call->to);
	/* BKO: initialize empty route list */
	list_init (&new_call->route_list);
}

/* internal constructor ! do not use !*/
OsipDialog *
osip_dialog_alloc (OsipUA * ua)
{
	OsipDialog *new_call;

	new_call = smalloc (sizeof (OsipDialog));
	osip_dialog_init (new_call, ua);
	return new_call;
}

/* this is called by the application to create a new call-leg */

/**
 * osip_dialog_new:
 * @ua: an existing user agent.
 *
 * Creates a new #OsipDialog object managed by the user agent @ua.
 *
 * Returns: a new call leg.
 */
OsipDialog *
osip_dialog_new (OsipUA * ua)
{
	OsipDialog *new_call;
	new_call = osip_dialog_alloc (ua);
	osip_dialog_init2 (new_call, ua);
	return (new_call);
}



/* create a new call-leg from an initial incoming request . The function
assumes that there is no existing call-leg that match the new request.
You should use osip_dialog_find() to check this */

/**
 * osip_dialog_new_from_incoming_trn:
 * @trn: an incoming transaction
 *
 * Create a new call-leg from an initial incoming request. The function
 * assumes that there is no existing call-leg that match the new request.
 * You should use #osip_dialog_find() to check this.
 *
 * Returns: a new call leg.
 */
OsipDialog *
osip_dialog_new_from_incoming_trn (transaction_t * trn)
{
	OsipDialog *call;
	OsipUA *ua;
	sip_t *sipmsg;
	contact_t *ctt;
	char *from;
	char *to;
	int i;

	call = ua_transaction_get_dialog (trn);
	if (call != NULL)
	{
		osip_trace (OSIP_INFO1,
			    ("A call-leg already exists for this transaction. (%x)",
			     trn));
		return NULL;
	}

	/* first find a ua for this call-leg */
	ua = osip_ua_find (trn->orig_request);

	i = from_2char (trn->orig_request->from, &from);
	if (i != 0)
		return NULL;
	osip_trace (OSIP_INFO1,
		    ("%s has called at %i.\n", from, time (NULL)));
	sfree (from);


	if (ua == NULL)
	{
		i = to_2char (trn->orig_request->to, &to);
		if (i != 0)
			return NULL;
		/* we should return a 404 Not Found */
		osip_trace (OSIP_INFO1,
			    ("error: Requested user (%s) does not exist.\n",
			     to));
		sfree (to);
		/* take the first user agent of the list to respond to this request  :  Maybe revisit */
		respond_to_request (def_manager->config, trn, 404);
		return NULL;
	}
	/* check if the user agent can answer to the call */
	if (ua->dialog_count >= ua->max_dialogs)
	{
		/*respond a 486 busy here */
		respond_to_request (ua->config, trn, 486);
		return NULL;
	}
	/* the transaction is accepted, it can open a new dialog*/
	
	call = osip_dialog_alloc (ua);
	/* then setup the dialog according the initial incoming request */
	sipmsg = trn->orig_request;
	ua_transaction_incoming_set_dialog (trn, call);
	transaction_set_out_socket(trn,ua->manager->send_sock);
	if (MSG_IS_INVITE (sipmsg))
		osip_dialog_set_state (call, DIALOG_INVITED);
	else if (MSG_IS_REGISTER (sipmsg))
		osip_dialog_set_state (call, DIALOG_FAKE);
	else
	{
		osip_trace (OSIP_WARNING,
			    ("Unsupported new incoming request."));
		osip_dialog_release (call);
		return NULL;
	}
	call_id_clone (sipmsg->call_id, &call->callid);
	msg_getcontact (sipmsg, 0, &ctt);
	if (ctt != NULL)
	{
		/* BKO: only set contact if we have not done so already */
		if (call->peer == NULL)
		{
			contact_clone (ctt, &call->peer);			
		}
	}
	else
	{
		osip_trace (OSIP_WARNING,
			    ("warning: No contact field in incoming request.\n"));
	}
	from_clone (sipmsg->from, &call->to);	/* The from from the incoming request
						 * will become the to the future outgoing request (including tag) */
	call->remote_cseq = atoi (sipmsg->cseq->number);
	return call;
}

/**
 * osip_dialog_update_from_response:
 * @call: a call leg.
 * @resp: a sip response.
 *
 * This function should be called each time a new response is received with
 * a call-leg. It updates the context of the call-leg according to the
 * response.
 *
 */
void
osip_dialog_update_from_response (OsipDialog * call, sip_t * resp)
{
	if (!call->tag_set)
	{
		/* if the call-leg is new, it has no tag parameter in to. If the server (maybe a proxy) give us one (it should)
		 * then we must remember it */
		from_copy_tag ((from_t *) call->to, (from_t *) resp->to);
		call->tag_set = 1;
	}
	/* BKO: if there is a record-route list here, save it so we
	 * can use it */
	{
		record_route_t *rr;
		route_t *rr2;
		int pos;
		int i;
		char *tmp;

		if (!list_eol (resp->record_routes, 0))
		{
			for (pos = resp->record_routes->nb_elt - 1; pos >= 0;
			     pos--)
			{
				rr = (record_route_t *) list_get (resp->
								  record_routes,
								  pos);
				i = record_route_2char (rr, &tmp);
/*		    	printf("********** route: %s\n", tmp); */
				route_init (&rr2);
				route_parse (rr2, tmp);
				list_add (&call->route_list, rr2, 0);
				sfree (tmp);
			}
		}
	}

	/* BKO: if there is a contact, append to the route list */
	{
		contact_t *ctt;
		char *tmp;
		int i;
		route_t *rr2;
		msg_getcontact (resp, 0, &ctt);
		if (ctt != NULL)
		{
			osip_trace (OSIP_INFO1,
				    ("Contact in 2xx - adding to route list\n"));
			i = contact_2char (ctt, &tmp);
			route_init (&rr2);
			route_parse (rr2, tmp);
			list_add (&call->route_list, rr2, -1);
		}
	}


	if (call->peer == NULL)
	{
		/* BKO: set to the first item from the route list */
		if (!list_eol (&call->route_list, 0))
		{
			contact_t *p;
			char *tmp;

			route_t *route =
				(route_t *) list_get (&call->route_list, 0);
			/* must also pop the route */
			list_remove (&call->route_list, 0);

			route_2char (route, &tmp);

			contact_init (&p);
			contact_parse (p, tmp);

			call->peer = p;
		}
	}
}

/* update from incoming request */

void osip_dialog_update_from_request(OsipDialog *dia, sip_t *sipmsg)
{
	cseq_t *cseq=msg_getcseq(sipmsg);
	char *number=cseq_getnumber(cseq);
	dia->remote_cseq=atoi(number);
}

/* accept an invitation for an existing call-leg*/
/**
 *osip_dialog_accept_invite:
 *@dialog: a dialog.
 *@trn:	   a transaction initiated by INVITE from a remote client.
 *
 * On reception of the 200 Ok for invite, then the osipua library calls
 * the "INVITE" signal handler. Then the user can choose to accept the call
 * using this function.
 * If the user wants to reject the call, then it should call
 * osip_dialog_bye().
 *
 */
void
osip_dialog_accept_invite (OsipDialog * dialog, transaction_t * trn)
{
	OsipUA *ua;
	
	/* check if we have good status for this call to answer 200OK */
	if (osip_dialog_exists (dialog) == 0)
	{
		osip_trace (OSIP_INFO1,
			    ("error: SIP dialog does not exist.\n"));
		return;
	}
	ua = osip_dialog_get_ua (dialog);


	if (trn == NULL)
	{
		osip_trace (OSIP_INFO1,
			    ("error: could not find last invite incoming transaction\n"));
		return;
	}
	osip_dialog_set_state (dialog, DIALOG_ESTABLISHED);
	osip_dialog_respond (dialog, trn, 200);
	
}

/* register your identity to a registrar */

/**
 * osip_dialog_register:
 * @call_leg: a call-leg.
 * @context: a registration context.
 *
 * Sends a REGISTER request to a registrar.
 *
 * Returns: 0 if the REGISTER was sent, a negative value if not.
 */
int
osip_dialog_register (OsipDialog * call_leg, RegistrationCtxt * context)
{
	sip_t *sipmesg;
	char *tmp;
	url_t *rgt;
	from_t *addr_of_rec;
	int err;
	OsipUA *ua=osip_dialog_get_ua(call_leg);
	
	if (context->registrar == NULL)
	{
		osip_trace (OSIP_INFO1,
			    ("error: You have to specify a registrar.\n"));
		return (-1);
	}

	url_init (&rgt);
	err = url_parse (rgt, context->registrar);
	if (err != 0)
	{
		osip_trace (OSIP_INFO1,
			    ("error: Bad registrar address (%s).\n",
			     context->registrar));
		sfree (rgt);
		return -EINVAL;
	}

	if (context->address_of_record != NULL)
	{
		from_init (&addr_of_rec);
		err = from_parse (addr_of_rec, context->address_of_record);
		if (err != 0)
		{
			osip_trace (OSIP_INFO1,
				    ("error: Bad address of record (%s).\n",
				     context->address_of_record));
			sfree (addr_of_rec);
			return -EINVAL;
		}
		/* set to=from= address_of_record */
		from_free (call_leg->from);
		sfree (call_leg->from);
		call_leg->from = addr_of_rec;
		/* add the ua tag to the from */
		from_set_tag(call_leg->from,sgetcopy(ua->fromtag));
		from_parse (call_leg->to, context->address_of_record);
	}
	else
	{
		/* if no address of record is given in the context, then use the default from field of the call-leg */
		to_2char ((from_t *) call_leg->from, &tmp);
		to_parse (call_leg->to, tmp);
		sfree (tmp);
	}
	/* remove the tag in the to feild! */
	if (!list_eol (call_leg->to->gen_params, 0))
	{
		generic_param_t *gp = list_get (call_leg->to->gen_params, 0);
		list_remove (call_leg->to->gen_params, 0);
		generic_param_free (gp);
		sfree (gp);
	}
	call_leg->reg_context = context;
	context->ref_count++;	/* context is in use ! */
	call_leg->local_cseq++;
	sipmesg = osip_dialog_compose_request (call_leg, REGISTER);
	if (sipmesg == NULL)
		return (-1);
	osip_dialog_send_request (call_leg, sipmesg);
	//osip_dialog_set_state (call_leg, CALL_REGISTERING);
	return 0;
}

int
osip_dialog_register_with_authentication (OsipDialog * call_leg,
					  sip_t * previous_answer,char *passwd)
{
	sip_t *sipmesg;
	authorization_t *aut = NULL;
	proxy_authorization_t *proxy_aut = NULL;
	int i;
	RegistrationCtxt *regctx;
	char *uri;
	char *pass;
	OsipUA *ua=osip_dialog_get_ua(call_leg);

	/* set action */
	call_leg->local_cseq++;
	sipmesg = osip_dialog_compose_request (call_leg, REGISTER);
	if (sipmesg == NULL)
		return (-1);

	/* ADD AUTHENTICATION STUFF.... */
	url_2char (sipmesg->strtline->rquri, &uri);
	/*     if (MSG_TEST_CODE(previous_answer,407)
	 * ||MSG_TEST_CODE(previous_answer,401))
	 * { */
	if ((previous_answer->proxy_authenticate != NULL)
	    && (previous_answer->www_authenticate != NULL))
		return -1;
	if (passwd==NULL){
		/* then tries to find a password in the registration context */
		regctx=call_leg->reg_context;
		if (regctx!=NULL){
			pass=regctx->password;
		}else {
			osip_trace(OSIP_ERROR,("osip_dialog_register_with_authentification: no password, aborting"));
			sfree(uri);
			msg_free(sipmesg);
			sfree(sipmesg);
			return -1;
		}
	}else pass=passwd;
	
	if (previous_answer->proxy_authenticate != NULL)
	{
		i = osip_create_proxy_authorization_header (previous_answer,
							    uri,ua->contact->url->username,pass, &proxy_aut);
		if (i != 0)
		{
			osip_trace (OSIP_INFO1,
				    ("error: could not create authorization header.\n"));
			return -1;
		}
	}

	if (previous_answer->www_authenticate != NULL)
	{
		i = osip_create_authorization_header (previous_answer,
						      uri, ua->contact->url->username,pass, &aut);
		if (i != 0)
		{
			osip_trace (OSIP_INFO1,
				    ("error: could not build the proxy_authorization header.\n"));
			return -1;
		}
	}
	/* TODO */
	/* here we should copy the proxy_authorization header
	 * from the previous_answer to the new request. This
	 * proxy_authorization is the response given by the
	 * first proxy made for the second proxy!
	 * oSIP must be fixed here to support multiple
	 * proxy_authorization, authorization, www_authenticate,
	 * and proxy_authenticate in one message.
	 * jack@atosc.org
	 */
	/* !TODO */


	sipmesg->authorization = aut;
	if (proxy_aut != NULL)
		list_add (sipmesg->proxy_authorizations, proxy_aut, -1);
	/* ADD AUTHENTICATION STUFF.... */

	osip_dialog_send_request (call_leg, sipmesg);
	//osip_dialog_set_state (call_leg, CALL_REGISTERING);
	return 0;
}

/**
 * osip_dialog_unregister:
 * @call: a call-leg.
 * @ctxt: a registration context that has been used to make a previous registration.
 *
 * Sends a REGISTER request to a registrar with a expire=0, so that
 * the registration is cancelled.
 * Warning: the ctxt structure may be changed. Do not rely on it to make a new registration.
 *
 * Returns: 0 if the REGISTER was sent, a negative value if not.
 */
int
osip_dialog_unregister (OsipDialog * call, RegistrationCtxt * ctxt)
{
	int error;
	ctxt->expires = 0;
	error = osip_dialog_register (call, ctxt);
	return error;
}

/* invite somebody */
/**
 * osip_dialog_invite:
 * @call_leg: a call-leg.
 * @callee: the URI of the person to invite.
 *
 * Sends an invite to @callee.
 *
 * Returns: 0 if the invite was sent, a negative value if not.
 */
int
osip_dialog_invite (OsipDialog * call_leg, char *callee, char *body_mime)
{
	OsipUA *ua = osip_dialog_get_ua (call_leg);
	sip_t *sipmesg;
	char *callee_addr;
	int err;

	if (call_leg->status != DIALOG_NEW)
		return (-ESRCH);

	/* set to */
	callee_addr = sgetcopy (callee);	/* useful because we write on calle_addr */
	err = from_parse ((from_t *) (call_leg->to), callee_addr);
	sfree (callee_addr);
	if (err == SIP_SYNTAX_ERROR)
	{
		osip_trace (OSIP_ERROR, ("Invalid sip address.\n"));
		return -EINVAL;
	}

	/* compose an invite */
	call_leg->local_cseq++;
	sipmesg = osip_dialog_compose_request (call_leg, INVITE);
	if (sipmesg == NULL)
	{
		osip_trace (OSIP_INFO1,
			    ("error: could not build sip message.\n"));
		return (-1);
	}
	/* add the body */
	if (body_mime!=NULL)
		osip_dialog_add_body_to_request (call_leg, sipmesg, body_mime);
	/* and send it */
	osip_dialog_send_request (call_leg, sipmesg);
	osip_dialog_set_state (call_leg, DIALOG_INVITING);
	return (0);
}

/* invite somebody */
int
osip_dialog_reinvite_with_authentication (OsipDialog * call_leg,
					sip_t * previous_answer,char *password)
{
	OsipUA *ua = osip_dialog_get_ua (call_leg);
	sip_t *sipmesg;
	char *uri;
	authorization_t *aut = NULL;
	proxy_authorization_t *proxy_aut = NULL;


	call_leg->local_cseq++;
	sipmesg = osip_dialog_compose_request (call_leg, INVITE);
	if (sipmesg == NULL)
	{
		osip_trace (OSIP_INFO1,
			    ("error: could not build sip message.\n"));
		return (-1);
	}


	/* ADD AUTHENTICATION STUFF.... */
	url_2char (sipmesg->strtline->rquri, &uri);
	/*     if (MSG_TEST_CODE(previous_answer,407)
	 * ||MSG_TEST_CODE(previous_answer,401))
	 * { */
	if ((previous_answer->proxy_authenticate != NULL)
	    && (previous_answer->www_authenticate != NULL))
		return -1;
	if (previous_answer->proxy_authenticate != NULL)
	{
		int i = osip_create_proxy_authorization_header (previous_answer,
								uri,
								ua->contact->url->username,
								password,
								&proxy_aut);
		if (i != 0)
		{
			osip_trace (OSIP_INFO1,
				    ("error: could not build the authorization header.\n"));
			return -1;
		}
	}

	if (previous_answer->www_authenticate != NULL)
	{
		int i = osip_create_authorization_header (previous_answer,
							  uri,ua->contact->url->username,
								password, &aut);
		if (i != 0)
		{
			osip_trace (OSIP_INFO1,
				    ("error: could not build the proxy_authorization header.\n"));
			return -1;
		}
	}
	/* TODO */
	/* here we should copy the proxy_authorization header
	 * from the previous_answer to the new request. This
	 * proxy_authorization is the response given by the
	 * first proxy made for the second proxy!
	 * oSIP must be fixed here to support multiple
	 * proxy_authorization, authorization, www_authenticate,
	 * and proxy_authenticate in one message.
	 * jack@atosc.org
	 */
	/* !TODO */


	sipmesg->authorization = aut;
	if (proxy_aut != NULL)
		list_add (sipmesg->proxy_authorizations, proxy_aut, -1);
	/* !ADD AUTHENTICATION STUFF.... */

	/* and send it */
	osip_dialog_send_request (call_leg, sipmesg);
	osip_dialog_set_state (call_leg, DIALOG_INVITING);
	return (0);

}


/* bye and destroy the call-leg*/

/**
 * osip_dialog_bye:
 * @call_leg: a call-leg.
 *
 *
 * Sends a bye and ends the call-leg.
 *
 * Returns: 0 if the bye was sent, a negative value if not.
 */
int
osip_dialog_bye (OsipDialog * call_leg)
{
	sip_t *sipmesg;

	if (call_leg->status != DIALOG_ESTABLISHED)
	{
		osip_trace (OSIP_INFO1,
			    ("warning: no dialog established. (use CANCEL instead)\n"));
	}
	call_leg->local_cseq++;
	sipmesg = osip_dialog_compose_request (call_leg, BYE);
	if (sipmesg == NULL)
		return (-1);	/* to be set... */
	osip_dialog_send_request (call_leg, sipmesg);
	osip_dialog_release (call_leg);
	return (0);
}




/* returns 1 if call leg exists*/
int
osip_dialog_exists (OsipDialog * call)
{
	int pos;
	OsipDialog *callleg;

	for (pos = 0; pos < callleg_list.nb_elt; pos++)
	{
		callleg = (OsipDialog *) list_get (&callleg_list, pos);
		if (callleg == call)
			return 1;
	}
	return 0;
}


/**
 * osip_dialog_find:
 * @sipmsg: an incoming request.
 * @dialog: the dialog as the return value.
 *
 * Tries to find a call-leg that may be concerned by the incoming request.
 *
 * Returns: 1 if a dialog was found, 0 if not found, -1 if the message has to be discarded
 *	and no dialog should be created after.
 */
int
osip_dialog_find (sip_t * sipmsg, OsipDialog **ret_dialog)
{
	int pos = 0;
	int cond = 0;
	int error=0;
	OsipDialog *dialog;
	
	*ret_dialog = NULL;

	/*find an existant call-leg that may match  this new invite */
	while (!list_eol (&callleg_list, pos) && (!cond))
	{
		dialog = (OsipDialog *) list_get (&callleg_list, pos);
		/* there are much more things to check: callid, from tags, to etc... */
		if (call_id_match (sipmsg->call_id, dialog->callid) == 0)
		{
			cond=1;
			if (MSG_IS_REQUEST (sipmsg))
			{
				if (MSG_IS_ACK (sipmsg))
				{
					if (atoi (sipmsg->cseq->number) < dialog->remote_cseq)	
					{
						osip_trace (OSIP_INFO1,
							    ("Ack cseq does not match !!!!!!!!!!!!!\n"));
						error=1;
					}
				}
				else
				{
					if (atoi (sipmsg->cseq->number) <= dialog->remote_cseq)
					{
						error=1;
						osip_trace (OSIP_INFO1,
							    ("Request Cseq does not match !!!!!!!!!!!!!\n"));
					}
				}
			}
			else
			{
				if (atoi (sipmsg->cseq->number) < dialog->local_cseq)
				{
					osip_trace (OSIP_INFO1,
						    ("info: Response Cseq does not match !!!!!!!!!!!!!\n"));
					error=1;
				}
			}
		}
		pos++;
	}
	if (error) return -1;
	if (cond)
	{
		osip_trace (OSIP_INFO1,
			    ("call-leg has been found!\n"));
		/* an existing call-leg was found */
		*ret_dialog=dialog;
		return 1;
	}
	else return 0;
}


/**
 * osip_dialog_send_response:
 * @call: a call-leg
 * @trn: a transaction
 * @resp: a sip response
 *
 *
 * Sends the response @resp that will be part of the transaction @trn,
 * trn being part of the call-leg @call.
 *
 */
void
osip_dialog_send_response (OsipDialog * call, transaction_t * trn,
			   sip_t * resp)
{
	osip_send_response (call->ua->manager, trn, resp);
}



/**
 *osip_dialog_make_response:
 *@call: a dialog.
 *@transaction : a transaction
 *@code: code of response.
 *
 *Returns: a sip response.
**/
sip_t * osip_dialog_make_response(OsipDialog *call, transaction_t *trn, int code)
{
	sip_t *response;
	OsipUA *ua = osip_dialog_get_ua (call);
	int pos;
	BodyContext *bh;

	if (  (response=msg_make_template_reply(trn->orig_request,code))==NULL)
	{
		osip_trace (OSIP_ERROR,
			    ("osip_dialog_respond: Error in msg_make_template_reply().\n"));
		return NULL;
	}

	{
		generic_param_t *gp;
		/* BKO: this should be To:, not From: */
/*			from_t *from;
			from=msg_getfrom(response);
			from_tag_add(from,sgetcopy(ua->fromtag));
*/
		/* we must add a to tag if it is not already present */
		to_t *to;
		to = msg_getto (response);
		to_get_tag(to,&gp);
		if (gp==NULL)
			to_tag_add (to, sgetcopy (ua->fromtag));
	}
	/* BKO: we must save and mark the peer as well as the
	 * route here, from the record-route.  */
	{
		sip_t *req = trn->orig_request;

		record_route_t *rr;
		route_t *rr2;
		int pos;
		int i;
		char *tmp;

		/* this should only be done the first time! */
		if (!list_eol (req->record_routes, 0) &&
		    list_eol (&call->route_list, 0))
		{
			for (pos = req->record_routes->nb_elt - 1; pos >= 0;
			     pos--)
			{
				rr = (record_route_t *) list_get (req->
								  record_routes,
								  pos);
				i = record_route_2char (rr, &tmp);
/*			    printf("**** route %d: %s\n", pos, tmp); */
				route_init (&rr2);
				route_parse (rr2, tmp);
				/* the route list is the reverse of the
				 * code for a response we receive
				 * vs. response we send */
				list_add (&call->route_list, rr2, -1);
				sfree (tmp);
			}

			/* BKO: if there is a contact, append to the
			 * route list */
			{
				contact_t *ctt;
				char *tmp;
				int i;
				route_t *rr2;
				msg_getcontact (req, 0, &ctt);
				if (ctt != NULL)
				{
					osip_trace (OSIP_INFO1,
						    ("info: Contact in 2xx - adding to route list\n"));
					i = contact_2char (ctt, &tmp);
					route_init (&rr2);
					route_parse (rr2, tmp);
					list_add (&call->route_list, rr2, -1);
				}
			}
		}
	}

	/* we must set the peer from the route list */
	if (call->peer == NULL)
	{
		/* set to the first item from the route list */
		if (!list_eol (&call->route_list, 0))
		{
			contact_t *p;
			char *tmp;

			route_t *route =
				(route_t *) list_get (&call->route_list, 0);
			/* must also pop the route */
			list_remove (&call->route_list, 0);

			route_2char (route, &tmp);

			contact_init (&p);
			contact_parse (p, tmp);

			call->peer = p;
		}
	}



	/* the user agent MUST set a contact header in last RFC */
	{
		char *tmp =
			(char *) smalloc (60 * sizeof (char) +
					  strlen (ua->contact->url->
						  username));
		sprintf (tmp, "<sip:%s@%s:%i>", ua->contact->url->username,
			 ua->ua_ip4addr, ua->ua_port);
		msg_setcontact (response, tmp);
	}
	
	if ((code==200) && 
		( (MSG_IS_INVITE(trn->orig_request)) || (MSG_IS_OPTIONS(trn->orig_request)))){

		/* if we send a 200 Ok for INVITE or OPTIONS, then invoke the body handlers to add their bodies if they want */
		pos = 0;
		while (!list_eol (&call->body_contexts, pos))
		{	
			bh = (BodyContext *) list_get (&call->body_contexts, pos);
			body_context_gen_out_response (bh, response);
			pos++;
		}
	}
	return response;
}

/* respond to a request that is part from a call-leg : ex 200OK (INVITE)*/
/**
 * osip_dialog_respond:
 * @call: a call-leg.
 * @code: the code of the response.
 *
 *
 * This function creates a sip response in the context of the call-leg @call
 * for the last transaction and sends it automatically.
 * A sdp body can be optionnaly appended to the response.
 *
 */
void osip_dialog_respond(OsipDialog *call, transaction_t *transaction, int code)
{
	sip_t *response=osip_dialog_make_response(call,transaction,code);
	if (response!=NULL)
		osip_dialog_send_response (call, transaction, response);
}

/**
 * osip_dialog_ack:
 * @call: a call-leg.
 * @trn: a transaction part of the dialog.
 *
 *
 * This function generates and sends an ACK request that will end the
 * transaction @trn. Must never be called outside the osip callbacks.
 *
 */
void
osip_dialog_ack (OsipDialog * call, transaction_t * trn)
{
	sip_t *ack;
	char *dest;
	OsipUA *ua = osip_dialog_get_ua (call);
	int port = 0;

	ack = osip_dialog_compose_request (call, ACK);
	/* in acks, the request uri must be the same as in the request of transaction, unless call->peer exists */
	if (ack->strtline->rquri->host == NULL)
	{
		char *tmp = NULL;
		url_2char (trn->orig_request->strtline->rquri, &tmp);
		url_parse (ack->strtline->rquri, tmp);
		sfree (tmp);
	}
	/* see if we need a name resolution */
	ua_transaction_get_destination(trn,&dest,&port);
	if (inet_addr (dest) == -1)
	{
		async_resolv_and_send_ack(ua->manager,dest,port,ack);
		return;
	}
	/* ack for invite cannot be sent in the same way */
	udp_send (trn,ack,dest,port, trn->out_socket);

}

/**
 * osip_dialog_compose_request:
 * @call_leg: a call-leg.
 * @msgtype: the request type in INVITE, BYE, REGISTER.
 *
 * Creates a sip request of type @msgtype according to the context of
 * the call-leg @call_leg. The request is not sent.
 *
 * Returns: a sip message.
 */
sip_t *
osip_dialog_compose_request (OsipDialog * call_leg, int msgtype)
{
	sip_t *sipmsg;
	char *tmp;
	OsipUA *ua = (OsipUA *) call_leg->ua;
	RegistrationCtxt *context = call_leg->reg_context;
	int pos = 0;

	msg_init (&sipmsg);

	/* set to header */
	{			/* mandatory */
		to_2char (call_leg->to, &tmp);
		msg_setto (sipmsg, tmp);
		sfree (tmp);
	}
	/*set from header */
	{
		from_2char (call_leg->from, &tmp);
		msg_setfrom (sipmsg, tmp);
		sfree (tmp);
	}
	/* set via header: use ip4addr instead of name */
	{
		char *via = (char *) smalloc (60 * sizeof (char));
		sprintf (via, "SIP/2.0/UDP %s:%i;branch=%u", ua->ua_ip4addr,
			 ua->ua_port,new_random_number());
		msg_setvia (sipmsg, via);
		sfree (via);
	}
	/*set contact header */
	contact_2char (ua->contact, &tmp);
	msg_setcontact (sipmsg, tmp);
	sfree (tmp);
	if (msgtype == REGISTER)
	{
		/* add a tag "action" */
		contact_t *ctt;
		if (context == NULL)
		{
			osip_trace (OSIP_INFO1,
				    ("registration context is null!\n"))
				return NULL;
		}
		msg_getcontact (sipmsg, 0, &ctt);
		if (context->action == REGISTER_ACTION_REDIRECT)
			from_param_add (((from_t *) ctt), sgetcopy ("action"),
					sgetcopy ("redirect"));
		else if (context->action == REGISTER_ACTION_PROXY)
			from_param_add (((from_t *) ctt), sgetcopy ("action"),
					sgetcopy ("proxy"));
		tmp = smalloc (10);
		sprintf (tmp, "%i", context->expires);
		msg_setexpires (sipmsg, tmp);
		sfree (tmp);
		/* set the from url same as the address of record, because some server don't like other addresses */
		//sfree(sipmsg->from->url->host);
		//sipmsg->from->url->host=sgetcopy(sipmsg->to->url->host);
		/* remove the tag of the from field, because some server don't like it */

		//list_remove(sipmsg->from->gen_params,0);

	}

	/* set first line */
	sipmsg->strtline = (startline_t *) smalloc (sizeof (startline_t));
	sipmsg->strtline->sipmethod = (char *) smalloc (14);
	switch (msgtype)
	{
	case INVITE:
		strcpy (sipmsg->strtline->sipmethod, "INVITE");
		break;
	case BYE:
		strcpy (sipmsg->strtline->sipmethod, "BYE");
		break;
	case REGISTER:
		strcpy (sipmsg->strtline->sipmethod, "REGISTER");
		break;
	case ACK:
		strcpy (sipmsg->strtline->sipmethod, "ACK");
		break;
	}

	sipmsg->strtline->sipversion = (char *) smalloc (8 * sizeof (char));
	sprintf (sipmsg->strtline->sipversion, "SIP/2.0");
	sipmsg->strtline->statuscode = NULL;
	sipmsg->strtline->reasonphrase = NULL;
	url_init (&sipmsg->strtline->rquri);

	/* set request URI, depends on message type */
	switch (msgtype)
	{
	case REGISTER:
		url_parse (sipmsg->strtline->rquri, context->registrar);
		/* set also the expire field */
		/* set expires */

		break;
	case ACK:
		/* in acks, the request uri must be the same as in the request of transaction
		 * unless there is a OsipDialog::peer set by previous response */
		/* but we don't have the transaction reference here so we leave blank the
		 * request uri to be set in osip_dialog_ack() */

		if (call_leg->peer != NULL)
		{
			url_2char (call_leg->peer->url, &tmp);
			url_parse (sipmsg->strtline->rquri, tmp);
			sfree (tmp);
		}
		break;
	default:
		/*
		 * GRS
		 * we need to do the same as above for subsequent messages
		 * in the dialog (e.g. BYE)
		 */
		if (call_leg->peer == NULL)
			url_2char (call_leg->to->url, &tmp);
		else
			url_2char (call_leg->peer->url, &tmp);
		url_parse (sipmsg->strtline->rquri, tmp);
		sfree (tmp);
	}

	/* remove unwanted information... They do not appear on request URI */
	/* to do */

	/* set user agent information */
	msg_setuser_agent (sipmsg, "oSIP/Linphone-" VERSION);

	/* set call-id */
	{
		char *callid;

		call_id_2char (call_leg->callid, &callid);
		msg_setcall_id (sipmsg, callid);
		sfree (callid);
	}
	/* set cseq */
	{
		char *cseq = (char *) smalloc (31 * sizeof (char));
		snprintf (cseq, 30, "%i %s", call_leg->local_cseq,
			  sipmsg->strtline->sipmethod);
		msg_setcseq (sipmsg, cseq);
		sfree (cseq);
	}
	/* maybe need to set a subject ... */
	/* to do */

	/* now invoke the BodyContext so that they had their bodies to the sip message */
	pos = 0;
	while (!list_eol (&call_leg->body_contexts, pos))
	{
		BodyContext *bh;
		bh = list_get (&call_leg->body_contexts, pos);
		body_context_gen_out_request (bh, sipmsg);
		pos++;
	}

	return (sipmsg);
}

int
osip_dialog_add_body_to_request (OsipDialog * call, sip_t * msg,
				 char *body_mime)
{
	OsipUA *ua = osip_dialog_get_ua (call);
	BodyHandler *bh;
	BodyContext *bc;

	bc = osip_dialog_get_body_context (call, body_mime, 0);
	if (bc == NULL)
	{
		bh = osip_ua_find_handler (ua, body_mime);
		if (bh == NULL)
		{
			osip_trace (OSIP_WARNING,
				    ("There is no handler for %s\n",
				     body_mime));
			return -1;
		}
		bc = body_handler_create_context (bh, call);
		if (bc == NULL)
		{
			osip_trace (OSIP_ERROR,
				    ("The handler for %s could not create a new context\n",
				     body_mime));
			return -1;
		}
		osip_dialog_add_body_context(call,bc);
	}
	body_context_gen_out_request (bc, msg);
	return 0;
}

/**
 * osip_dialog_send_request:
 * @call_leg: a call-leg.
 * @sipmsg: a sip request.
 *
 *
 * Sends a sip request. The transaction is created to wrap the request and
 * to the list of transaction of call-leg @call_leg.
 *
 * Returns: 0 if successfull.
 */
int
osip_dialog_send_request (OsipDialog * call_leg, sip_t * sipmsg)
{
	sipevent_t *sipevent;
	transaction_t *transaction;
	char *dest;
	int port = 0;
	OsipUA *ua = (OsipUA *) call_leg->ua;

	if (!MSG_IS_ACK (sipmsg))
	{
		sipevent = osip_new_outgoing_sipmessage (sipmsg);
		transaction = ua_transaction_new (call_leg, sipmsg);
	}
	else
	{
		osip_trace (OSIP_INFO1,
			    ("error: you must not use osip_dialog_send_request() to send acks."));
		/* "Acks must only be sent using osip_dialog_ack()" */
		exit (1);
	}
	sipevent->transactionid = transaction->transactionid;
	/* see if we need a name resolution */

	ua_transaction_get_destination (transaction, &dest, &port);
	if (inet_addr (dest) == -1)
		return (async_resolv_and_send(ua->manager, transaction, sipevent));

	ua_transaction_execute (transaction, sipevent);
	return transaction->transactionid;
}

/**
 * osip_dialog_release:
 * @call: a call-leg.
 *
 *
 * Schedule the end of the call-leg. The call-leg object will be destroyed as
 * soon as it the osip stack will terminate its last transaction.
 *
 */


/* Destroy a call-leg and remove it from ua list and stack list*/

/**
 * osip_dialog_destroy:
 * @call: a call-leg.
 *
 *
 * Destroy a call-leg and all data associated with it.
 * BUT: this function should not be called directly. You should use
 * #osip_dialog_release instead to schedule the destruction of the call
 * leg.
 *
 * Returns 0.
 */

int
osip_dialog_destroy (OsipDialog * call)
{
	int err = 0;
	struct _OsipUA *ua = call->ua;

	err = list_remove_el (&callleg_list, call);
	if (err < 0)
		osip_trace (OSIP_ERROR,
			    ("Could not remove dialog from list.\n"));
	err = list_remove_el (&ua->call_list, call);
	
	if (err < 0){
		osip_trace (OSIP_ERROR,("Could not remove dialog from ua list.\n"));
	}
	else {
		osip_trace (OSIP_INFO1,
		    ("Call leg is removed. It remains %i call legs in the ua list.\n",
		     list_size (&ua->call_list)));
	}
	
	if (call->to != NULL)
	{
		from_free (call->to);
		sfree (call->to);
	}
	if (call->from != NULL)
	{
		from_free (call->from);
		sfree (call->from);
	}
	if (call->callid != NULL)
	{
		call_id_free (call->callid);
		sfree (call->callid);
	}
	if (call->registrar != NULL)
	{
		url_free (call->registrar);
		sfree (call->registrar);
	}
	sfree (call);
	return (err);
}



/**
 * osip_dialog_release:
 * @call: a dialog.
 *
 * Release a dialog. You can't use it anymore, and the object will be
 * destroyed by the stack when the last transaction of the call-leg
 * will terminate, either by successfull completion or timeout.
 *
 *
 */
void osip_dialog_release(OsipDialog *call)
{
	OsipUA *ua=osip_dialog_get_ua(call);
	ua->dialog_count--;
	osip_dialog_set_state(call,DIALOG_TERMINATED);
}

void
osip_dialog_add_body_context (OsipDialog * call, BodyContext * handler)
{
	if (handler == NULL)
	{
		osip_trace (OSIP_INFO1,
			    ("Cannot add NULL body context to dialog.\n"));
		return;
	}
	list_add (&call->body_contexts, (void *) handler, 0);
}


BodyContext *
osip_dialog_get_body_context (OsipDialog * call, char *body_mime, int pos)
{
	BodyContext *context = NULL;
	char *tmp;
	int i = 0;
	int body_pos = 0;	/* the position in list of the body handler of the same type */
	//printf("entering osip_dialog_get_body_context list_size=%i\n", list_size(&call->body_contexts));
	while (!list_eol (&call->body_contexts, i))
	{
		context = (BodyContext *) list_get (&call->body_contexts, i);
		tmp = body_context_get_mime (context);
		//printf("Comparing %s <> %s , context=%x, handler=%x\n",tmp,body_mime,context,context->handler);
		if (strcmp (tmp, body_mime) == 0)
		{
			if ((body_pos == pos) || (pos == -1))
				return context;
			else
				body_pos++;
		}
		i++;
	}
	return NULL;
}
