/*
  The osipua library is a library based on oSIP that implements CallLeg and User Agent
  level.
  Copyright (C) 2001  Simon MORLAT simon.morlat@free.fr
  											Aymeric MOIZARD jack@atosc.org
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "osipua.h"
#include "utils.h"
#include "uatransaction.h"
#include "osipmanager.h"
#include <osip/smsg.h>
#include <osip/smsgtypes.h>

int ua_transaction_get_destination(transaction_t *trn, char **destination, int *port)
{
	if (trn->ctx_type==ICT){
		*destination=trn->ict_context->destination;
		*port=trn->ict_context->port;
	}
	else if (trn->ctx_type==NICT){
		*destination=trn->nict_context->destination;
		*port=trn->nict_context->port;
	}else {
		osip_trace(OSIP_ERROR,("Could not get destination for transaction: ctx_type=%i\n",trn->ctx_type));
		*destination=NULL;
		*port=0;
		return -1;
	}
	return 0;
}

void ua_transaction_set_destination(transaction_t *trn,char *destination, int port)
{
	if (trn->ctx_type==ICT){
		ict_set_destination(trn->ict_context,destination,port);
	}
	else if (trn->ctx_type==NICT){
		nict_set_destination(trn->nict_context,destination,port);
	}else {
		osip_trace(OSIP_ERROR,("Could not set destination for transaction: ctx_type=%i\n",trn->ctx_type));
	}
}

transaction_t * ua_transaction_new(OsipDialog *call,sip_t *sipmsg)
{
	transaction_t *transaction;
	OsipUA *ua;
	context_type_t type;
	int port=5060;
	
	if (call==NULL) return NULL;
	
	ua=(OsipUA*)call->ua;
	if (MSG_IS_INVITE(sipmsg))
	{
		type=ICT;
	}else type=NICT;
	transaction_init(&transaction,
			type,
	   		ua->config,
			sipmsg);
	transaction_set_out_socket(transaction,ua->manager->send_sock);
	/* if an outbound proxy is set for this call-leg, then set it to the transaction */
	if ((ua->registrar!=NULL) && (ua->flags & OSIP_UA_USE_PROXY) )
	{
		
		if (ua->registrar->port!=NULL) port=atoi(ua->registrar->port);
		ua_transaction_set_destination(transaction,sgetcopy(ua->registrar->host),port);
		
	}else{
		url_t *requri;
		/* else set the destination in the req uri */
		requri=msg_geturi(sipmsg);
		if (requri->port!=NULL) port=atoi(requri->port);
		ua_transaction_set_destination(transaction,sgetcopy(requri->host),port);
	}
	   		
	ua_transaction_outgoing_set_dialog(transaction,call);
	
	return transaction;
}

void ua_transaction_incoming_set_dialog(transaction_t *trn,OsipDialog *call)
{
	OsipUA *ua=osip_dialog_get_ua(call);
	/* add the transaction to the call-leg list of incoming transaction*/
	list_add(&call->incoming_transactions,trn,-1);
	trn->your_instance=call;
	transaction_set_out_socket(trn,ua->manager->send_sock);
}

void ua_transaction_outgoing_set_dialog(transaction_t *trn,OsipDialog *call)
{
	/* add the transaction to the call-leg list of outgoing transaction*/
	list_add(&call->outgoing_transactions,trn,-1);
	trn->your_instance=call;
}


void ua_transaction_free(transaction_t *transaction)
{
	OsipDialog *call;
	OsipUA *ua;
	int error;
	OsipManager *manager;
	
	/* check if the call_leg that the transaction is part of should not be deleted too*/
	call=ua_transaction_get_dialog(transaction);
	if (call==NULL)
	{
		osip_trace(OSIP_WARNING,("error: CallLeg==NULL ! This must not happen!"));
		return;
	}
	ua=osip_dialog_get_ua(call);
	manager=ua->manager;
	/* remove the transaction from the call leg list of transaction*/
	/* FIXME: we try to remove it on the two list, but we know it can
	be only in one of them */
	list_remove_el(&call->incoming_transactions,(void*)transaction);
	list_remove_el(&call->outgoing_transactions,(void*)transaction);
	
	/* now if it was the last active transaction of the dialog, delete the dialog*/
	if (list_size(&call->incoming_transactions)+list_size(&call->outgoing_transactions)==0)
	{
		switch (call->status)
		{
			case DIALOG_FAKE:
			case DIALOG_INVITING:
				/* if the transaction timed out, notify it */
				if (transaction->last_response==NULL){
				
					error=-ETIMEDOUT;
					/* the invite times out. The call-leg is lost. */
					if (ua->faillure!=NULL)
						ua->faillure(call,transaction,NULL,(void*)&error);
					/* then destroy the dialog*/
					osip_dialog_destroy(call);
				}
				
			break;
			case DIALOG_TERMINATED:
				osip_dialog_destroy(call);
			break;
		}
	}
	/* put the transaction in the free list */
	fifo_add(&manager->garbage_trn,(void*)transaction);
}

int is_in_recv_thread_context(OsipManager *manager)
{
	return (manager->thread_pid==getpid());
}

/* execute a transaction outside the receiving thread */
void ua_transaction_execute(transaction_t *trn, sipevent_t *ev)
{
	sipevent_t *se;
	OsipManager *manager;
	OsipDialog *dialog=ua_transaction_get_dialog(trn);
	OsipUA *ua;
	int cond;
	
	if (dialog==NULL){
		osip_trace(OSIP_WARNING,("ua_transaction_execute: could not get dialog transaction.\n"));
		manager=def_manager;
	}else {
		ua=osip_dialog_get_ua(dialog);
		manager=ua->manager;
	}
	/* check if we are within the process of the receiving thread 
	If you we are in, (probably in one of the osip callbacks), then we can safely run
	transaction_execute().
	If not the we have to take a lock to prevent the receiving thread from executing
	the transaction while we process it */
	cond=is_in_recv_thread_context(manager);
	
	if (cond){
		fifo_add(trn->transactionff,ev);
	}
	else{
		smutex_lock(manager->mutex);
		transaction_execute(trn,ev);
		smutex_unlock(manager->mutex);
	}
#if 0
	/* lock the receiving thread while processing */
	if (!cond) smutex_lock(manager->mutex);
	/* execute pending events before */
	while ( (se = (sipevent_t *)fifo_tryget(trn->transactionff)) !=NULL){
		transaction_execute(trn,se);
	}
	transaction_execute(trn,ev);
	if (!cond) smutex_unlock(manager->mutex);
#endif
	
	
}
