/*
  The oSIP library implements the Session Initiation Protocol (SIP -rfc2543-)
  Copyright (C) 2001  Aymeric MOIZARD jack@atosc.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "utils.h"
#include <osip/sdp.h>
#include "osipmanager.h"
#include "resolver.h"
#include "uatransaction.h"




void
respond_to_request (osip_t * config, transaction_t * transaction, int code)
{
	sip_t *response;
	response=msg_make_template_reply(transaction->orig_request,code);
	osip_send_response(def_manager,transaction,response);
}


void
osip_send_response (OsipManager * mgr, transaction_t * transaction,
		    sip_t * response)
{
	via_t *via;
	sipevent_t *sipevent;
	/* we must check if first via has x.x.x.x address. If not, we must resolve it */
	msg_getvia (response, 0, &via);
	if (via == NULL)
	{
		osip_trace (OSIP_INFO1,
			    ("error: Cannot send response because it has no via field"));
		return;
	}

	sipevent = osip_new_outgoing_sipmessage (response);
	sipevent->transactionid = transaction->transactionid;

	if (inet_addr (via->host) == -1)
	{
		/* need name resolution */
		async_resolv_and_send (def_manager, transaction, sipevent);
		return;
	}
	ua_transaction_execute(transaction,sipevent);
}


sip_t *
msg_make_template_reply (sip_t * request, int code)
{
	sip_t *response;
	char *tmp;
	int pos;
	
	msg_init (&response);
	msg_setversion (response, sgetcopy ("SIP/2.0"));
	tmp = smalloc (5);
	snprintf (tmp, 5, "%i", code);
	msg_setstatuscode (response, tmp);
	msg_setreasonphrase (response, msg_getreason (code));

	to_clone (request->to, &response->to);
	from_clone (request->from, &response->from);


	/* via headers */
	{
		pos = 0;
		while (!list_eol (request->vias, pos))
		{
			via_t *via;
			via = (via_t *) list_get (request->vias, pos);
			via_2char (via, &tmp);
			
			msg_setvia (response, tmp);
			sfree (tmp);
			pos++;
		}
	}

	call_id_clone(request->call_id,&response->call_id);
	
	cseq_clone(request->cseq,&response->cseq);
	
	/* if a record-route header is found, set a route a header
	 * with the reverse order of urls  and copy the record route
	 * to the response. */
	{
		record_route_t *rr, *rr2;
		contact_t *ct = NULL;
		/* this test is useles!!! */
		/* if (request->record_routes!=NULL) */
		if (!list_eol (request->record_routes, 0))
		{
			for (pos = request->record_routes->nb_elt - 1;
			     pos >= 0; pos--)
			{
				rr = (record_route_t *) list_get (request->
								  record_routes,
								  pos);
				record_route_2char (rr, &tmp);
				/* BKO: do not set the route in responses */
/* 				i=msg_setroute(response,tmp); */
				record_route_init (&rr2);
				record_route_parse (rr2, tmp);
				list_add (response->record_routes, rr2, 0);
				sfree (tmp);
			}
			/* BKO */
		}
	}

	/* set user agent information */
	msg_setuser_agent (response, "oSIP/Linphone-" VERSION);
	return response;
}




int
async_resolv_and_send (OsipManager * manager, transaction_t * transaction,
		       sipevent_t * sipevent)
{
	/* put the transaction,event on the fifo of the resolver thread */
	async_resolv_t *te;

	osip_trace (OSIP_INFO1, ("info: Name resolution requested.\n"));
	if (manager->resolv_run_cond == 0)
		osip_manager_start_resolver (manager);
	te = smalloc (sizeof (async_resolv_t));
	te->transaction = transaction;
	te->event = sipevent;
	te->dest = NULL;
	te->port = 0;
	te->ackmsg = NULL;
	fifo_add (manager->resolv_fifo, (void *) te);
	return transaction->transactionid;
}

int
async_resolv_and_send_ack (OsipManager * manager, char *dest, int port,
			   sip_t * ackmsg)
{
	/* put the transaction,event on the fifo of the resolver thread */
	async_resolv_t *te;

	osip_trace (OSIP_INFO1, ("info: Name resolution requested.\n"));
	if (manager->resolv_run_cond == 0)
		osip_manager_start_resolver (manager);
	te = smalloc (sizeof (async_resolv_t));
	te->transaction = NULL;
	te->event = NULL;
	te->dest = sgetcopy (dest);
	te->port = port;
	te->ackmsg = ackmsg;
	fifo_add (manager->resolv_fifo, (void *) te);
	return 0;
}



/*search element el in list, and remove it*/
int
list_remove_el (list_t * list, void *el)
{
	int pos = 0;
	int cond = 0;
	node_t *nd;

	nd = list->node;
	while ((pos < list->nb_elt) && (!cond))
	{
		if (nd->element == el)
			cond = 1;
		else
		{
			pos++;
			nd = nd->next;
		}
	}
	if (cond)
		list_remove (list, pos);
	else
		return (-1);
	return (0);
}

void
callid_getcopy (call_id_t * dest, call_id_t * source)
{
	dest->number = smalloc (strlen (source->number) + 1);
	strcpy (dest->number, source->number);
	dest->host = smalloc (strlen (source->host) + 1);
	strcpy (dest->host, source->host);
}

void
list_free (list_t * li, void (*destructor) (void *))
{
	int pos = 0;
	void *elem;
	while (!list_eol (li, pos))
	{
		elem = list_get (li, pos);
		list_remove (li, pos);
		if (destructor != NULL)
			destructor (elem);
	}
}

void *
srealloc (void *p, int i)
{
	return realloc (p, i);
}

char *
make_message (const char *fmt, ...)
{
	/* Guess we need no more than 100 bytes. */
	int n, size = 100;
	char *p;
	va_list ap;
	if ((p = smalloc (size)) == NULL)
		return NULL;
	while (1)
	{
		/* Try to print in the allocated space. */
		va_start (ap, fmt);
		n = vsnprintf (p, size, fmt, ap);
		va_end (ap);
		/* If that worked, return the string. */
		if (n > -1 && n < size)
			return p;
		//printf("Reallocing space.\n");
		/* Else try again with more space. */
		if (n > -1)	/* glibc 2.1 */
			size = n + 1;	/* precisely what is needed */
		else		/* glibc 2.0 */
			size *= 2;	/* twice the old size */
		if ((p = srealloc (p, size)) == NULL)
			return NULL;
	}
}

/* from2 has tag parameter that we should copy to from1 if it does not have one */
int
from_copy_tag (from_t * from1, from_t * from2)
{
	generic_param_t *param1 = NULL, *param2 = NULL;
	from_param_getbyname (from1, "tag", &param1);
	if (param1 == NULL)
	{
		from_param_getbyname (from2, "tag", &param2);
		if (param2 != NULL)
		{
			//printf("getting from tag field!!!\n");
			from_param_add (from1, sgetcopy ("tag"),
					sgetcopy (param2->gvalue));
		}
	}
	return 0;
}

int
_tag_init (char **tag, int length)
{
	char *newtag = smalloc ((length + 2) * sizeof (char));
	snprintf (newtag, length + 1, "%x", rand ());
	*tag = newtag;
	return 0;
}

/* this function check is *url has a sip: prefix. If it has not, then it add ones in a newly allocated string. If no correction is made, it just returns
the initial string */
char *
check_sipurl (char *url)
{
	char *newstr;
	from_t *from;
	int err;

	/* try to parse to see if the address in correct */
	from_init (&from);
	err = from_parse (from, url);
	from_free (from);
	sfree (from);
	if (err == 0)
		return url;	/* OK */


	if (strncmp (url, "sip:", 4) == 0)
		return NULL;
	if (strncmp (url, "<sip:", 5) == 0)
		return NULL;

	newstr = smalloc (strlen (url) + 5);
	strcpy (newstr, url);
	sclrspace (newstr);
	if (strncmp (newstr, "sip:", 4) == 0)
		return newstr;
	if (strncmp (newstr, "<sip:", 5) == 0)
		return newstr;
	/* if we are here, there is no "sip:" at the begginning, so add it */

	if (newstr[0] == '<')
		sprintf (newstr, "<sip:%s", url + 1);	/* do a <sip:machin@truc> */
	else
		sprintf (newstr, "sip:%s", url);	/* do a sip:machin@truc */
	sclrspace (newstr);
	/* try to parse to see if the address in correct */
	from_init (&from);
	err = from_parse (from, newstr);
	from_free (from);
	sfree (from);
	if (err == 0)
		return newstr;	/* OK */
	else
		return (NULL);
}


int
msg_getbody_by_mime (sip_t * sipmsg, char *mime, body_t ** body)
{
	int i = 0;
	body_t *tmpbody;
	while (1)
	{
		msg_getbody (sipmsg, i, &tmpbody);
		if (strcmp (tmpbody->content_type->type, mime) == 0)
		{
			*body = tmpbody;
			return 0;
		}
		i++;
	}
	*body = NULL;
	return -1;
}

char *
content_type_get_type (content_type_t * ct)
{
	char *p;
	int len = 0;
	if (ct->type != NULL)
		len += strlen (ct->type);
	else
	{
		return NULL;
	}
	if (ct->subtype != NULL)
		len += strlen (ct->subtype);
	else
		return NULL;
	p = smalloc (len + 2);
	sprintf (p, "%s/%s", ct->type, ct->subtype);
	return p;
}
